from dataclasses import dataclass
from enum import Enum

import google.protobuf.message
from nibiru_proto.proto.perp.v1 import event_pb2 as perp_events
from nibiru_proto.proto.spot.v1 import event_pb2 as spot_events
from nibiru_proto.proto.stablecoin import events_pb2 as stablecoin_events  # noqa
from nibiru_proto.proto.vpool.v1 import event_pb2 as vpool_events


class EventType(Enum):
    """
    The events enum type shows the type of events available to parse from the nibiruwebsocket object.
    """

    # Perp events
    PositionChangedEvent = perp_events.PositionChangedEvent
    PositionSettledEvent = perp_events.PositionSettledEvent
    PositionLiquidatedEvent = perp_events.PositionLiquidatedEvent
    FundingRateChangedEvent = perp_events.FundingRateChangedEvent

    # Vpool events
    ReserveSnapshotSavedEvent = vpool_events.ReserveSnapshotSavedEvent
    SwapOnVpoolEvent = vpool_events.SwapOnVpoolEvent
    MarkPriceChanged = vpool_events.MarkPriceChangedEvent

    # Spot event
    PoolJoinedEvent = spot_events.EventPoolJoined
    PoolCreatedEvent = spot_events.EventPoolCreated
    PoolExitedEvent = spot_events.EventPoolExited
    AssetsSwappedEvent = spot_events.EventAssetsSwapped

    # Stablecoin events
    TransferEvent = stablecoin_events.EventTransfer
    MintStableEvent = stablecoin_events.EventMintStable
    BurnStableEvent = stablecoin_events.EventBurnStable
    MintNIBIEvent = stablecoin_events.EventMintNIBI
    BurnNIBIEvent = stablecoin_events.EventBurnNIBI
    RecollateralizeEvent = stablecoin_events.EventRecollateralize
    BuybackEvent = stablecoin_events.EventBuyback

    # Bank
    Transfer = "transfer"

    # Staking
    Delegate = "delegate"
    Unbond = "unbond"
    Redelegate = "redelegate"

    def get_full_path(self):
        if isinstance(self.value, str):
            return self.value
        proto_message: google.protobuf.message.Message = self.value
        return str(proto_message.DESCRIPTOR.full_name)


@dataclass
class EventCaptured:
    event_type: EventType
    payload: dict



import dataclasses

from nibiru.msg.bank import MsgsBank
from nibiru.msg.perp import MsgsPerp
from nibiru.msg.spot import MsgsSpot
from nibiru.msg.staking import MsgsStaking


@dataclasses.dataclass
class MsgClient:
    """
    The 'MsgClient' exposes all available messages in the Nibiru Chain Python SDK.
    The class attributes of the client separate these messages by module.

    Attributes:
        bank: Methods for the Cosmos x/bank module.
        spot: Methods for the Nibiru Chain x/spot module.
        perp: Methods for the Nibiru Chain x/perp module
        staking: Methods for the Cosmos x/staking and x/distribution modules.
    """

    bank = MsgsBank
    spot = MsgsSpot
    perp = MsgsPerp
    staking = MsgsStaking


Msg = MsgClient()



import dataclasses

from nibiru_proto.proto.cosmos.distribution.v1beta1 import tx_pb2 as pb_distribution
from nibiru_proto.proto.cosmos.staking.v1beta1 import tx_pb2 as pb_staking

from nibiru.pytypes import Coin, PythonMsg


class MsgsStaking:
    """
    Messages for the x/staking and x/distribution modules.

    Methods:
    - delegate: Delegate tokens to a validator
    - undelegate: Undelegate tokens from a validator
    - withdraw_delegator_reward: Withdraw the reward from a validator
    """

    @staticmethod
    def delegate(
        delegator_address: str,
        validator_address: str,
        amount: float,
    ) -> 'MsgDelegate':
        """
        Delegate tokens to a validator

        Attributes:
            delegator_address: str
            validator_address: str
            amount: float

        Returns:
            MsgDelegate: PythonMsg for type 'cosmos.staking.v1beta1.MsgDelegate'
        """
        return MsgDelegate(
            delegator_address=delegator_address,
            validator_address=validator_address,
            amount=amount,
        )

    @staticmethod
    def undelegate(
        delegator_address: str,
        validator_address: str,
        amount: float,
    ) -> 'MsgUndelegate':
        """
        Undelegate tokens from a validator

        Attributes:
            delegator_address (str): Bech32 address for the delegator
            validator_address (str): Bech32 valoper address
            amount (float): Amount of unibi to undelegate.
        """
        return MsgUndelegate(
            delegator_address=delegator_address,
            validator_address=validator_address,
            amount=amount,
        )

    @staticmethod
    def withdraw_delegator_reward(
        delegator_address: str,
        validator_address: str,
    ) -> 'MsgWithdrawDelegatorReward':
        """TODO
        Withdraw the reward from a validator

        Attributes:
            delegator_address (str)
            validator_address (str)

        Returns:
            MsgWithdrawDelegatorReward: _description_
        """
        return MsgWithdrawDelegatorReward(
            delegator_address=delegator_address, validator_address=validator_address
        )


@dataclasses.dataclass
class MsgDelegate(PythonMsg):
    """
    Delegate tokens to a validator

    Attributes:
        delegator_address: str
        validator_address: str
        amount: float
    """

    delegator_address: str
    validator_address: str
    amount: float

    def to_pb(self) -> pb_staking.MsgDelegate:
        """
        Returns the Message as protobuf object.

        Returns:
            staking_pb.MsgDelegate: The proto object.

        """
        return pb_staking.MsgDelegate(
            delegator_address=self.delegator_address,
            validator_address=self.validator_address,
            amount=Coin(self.amount, "unibi")._generate_proto_object(),
        )


@dataclasses.dataclass
class MsgUndelegate(PythonMsg):
    """
    Undelegate tokens from a validator

    Attributes:
        delegator_address: str
        validator_address: str
        amount: float
    """

    delegator_address: str
    validator_address: str
    amount: float

    def to_pb(self) -> pb_staking.MsgUndelegate:
        """
        Returns the Message as protobuf object.

        Returns:
            staking_pb.MsgUndelegate: The proto object.

        """
        return pb_staking.MsgUndelegate(
            delegator_address=self.delegator_address,
            validator_address=self.validator_address,
            amount=Coin(self.amount, "unibi")._generate_proto_object(),
        )


@dataclasses.dataclass
class MsgWithdrawDelegatorReward(PythonMsg):
    """
    Withdraw the reward from a validator

    Attributes:
        delegator_address (str)
        validator_address (str)
    """

    delegator_address: str
    validator_address: str

    def to_pb(self) -> pb_distribution.MsgWithdrawDelegatorReward:
        """
        Returns the Message as protobuf object.

        Returns:
            tx_pb.MsgWithdrawDelegatorReward: The proto object.

        """
        return pb_distribution.MsgWithdrawDelegatorReward(
            delegator_address=self.delegator_address,
            validator_address=self.validator_address,
        )



# These import statements export the types to 'nibiru.pytypes'.

from nibiru.pytypes.common import (  # noqa # TODO move constants to a constants.py file.; noqa
    GAS_PRICE,
    MAX_MEMO_CHARACTERS,
    Coin,
    Direction,
    PoolAsset,
    PoolType,
    PythonMsg,
    Side,
    TxConfig,
    TxType,
)
from nibiru.pytypes.event import Event, RawEvent, TxLogEvents  # noqa
from nibiru.pytypes.network import Network, NetworkType  # noqa
from nibiru.pytypes.tx_resp import RawTxResp, TxResp  # noqa



import dataclasses
from typing import List, Union

from nibiru_proto.proto.spot.v1 import pool_pb2 as pool_tx_pb
from nibiru_proto.proto.spot.v1 import tx_pb2 as pb

from nibiru.pytypes import Coin, PoolAsset, PoolType, PythonMsg


class MsgsSpot:
    """MsgsSpot has methods for building messages for transactions on Nibi-Swap.

    Methods:
    - create_pool: Create a pool using the assets specified
    - exit_pool: Exit a pool using the specified pool shares
    - join_pool: Join a pool using the specified tokens
    - swap: Swap the assets provided for the denom specified
    """

    @staticmethod
    def create_pool(
        creator: str,
        swap_fee: float,
        exit_fee: float,
        a: int,
        pool_type: PoolType,
        assets: List[PoolAsset],
    ) -> 'MsgCreatePool':
        return MsgCreatePool(
            creator=creator,
            swap_fee=swap_fee,
            exit_fee=exit_fee,
            a=a,
            pool_type=pool_type,
            assets=assets,
        )

    @staticmethod
    def join_pool(
        sender: str,
        pool_id: int,
        tokens: Union[Coin, List[Coin]],
    ) -> 'MsgJoinPool':
        return MsgJoinPool(sender=sender, pool_id=pool_id, tokens=tokens)

    @staticmethod
    def exit_pool(
        sender: str,
        pool_id: int,
        pool_shares: Coin,
    ) -> 'MsgExitPool':
        return MsgExitPool(sender=sender, pool_id=pool_id, pool_shares=pool_shares)

    @staticmethod
    def swap(
        sender: str,
        pool_id: int,
        token_in: Coin,
        token_out_denom: str,
    ) -> 'MsgSwapAssets':
        return MsgSwapAssets(
            sender=sender,
            pool_id=pool_id,
            token_in=token_in,
            token_out_denom=token_out_denom,
        )


@dataclasses.dataclass
class MsgCreatePool(PythonMsg):
    """
    Create a pool using the assets specified

    Attributes:
        creator (str): The creator address
        swap_fee (float): The swap fee required for the pool
        exit_fee (float): The exit fee required for the pool
        assets (List[PoolAsset]): The assets to compose the pool
    """

    creator: str
    swap_fee: float
    exit_fee: float
    a: int
    pool_type: PoolType
    assets: List[PoolAsset]

    def to_pb(self) -> pb.MsgCreatePool:
        """
        Returns the Message as protobuf object.

        Returns:
            pb.MsgCreatePool: The proto object.

        """
        pool_assets = [
            pool_tx_pb.PoolAsset(
                token=a.token._generate_proto_object(), weight=str(int(a.weight * 1e6))
            )
            for a in self.assets
        ]

        swap_fee_dec = str(int(self.swap_fee * 1e18))
        exit_fee_dec = str(int(self.exit_fee * 1e18))

        return pb.MsgCreatePool(
            creator=self.creator,
            pool_params=pool_tx_pb.PoolParams(
                swap_fee=swap_fee_dec,
                exit_fee=exit_fee_dec,
                pool_type=self.pool_type,
                A=str(int(self.a)),
            ),
            pool_assets=pool_assets,
        )


@dataclasses.dataclass
class MsgJoinPool(PythonMsg):
    """
    Join a pool using the specified tokens

    Attributes:
        sender (str): The creator address
        pool_id (int): The id of the pool to join
        tokens (List[Coin]): The tokens to be bonded in the pool
    """

    sender: str
    pool_id: int
    tokens: Union[Coin, List[Coin]]

    def to_pb(self) -> pb.MsgJoinPool:
        """
        Returns the Message as protobuf object.

        Returns:
            pb.MsgJoinPool: The proto object.

        """
        if isinstance(self.tokens, Coin):
            self.tokens = [self.tokens]
        return pb.MsgJoinPool(
            sender=self.sender,
            pool_id=self.pool_id,
            tokens_in=[token._generate_proto_object() for token in self.tokens],
        )


@dataclasses.dataclass
class MsgExitPool(PythonMsg):
    """
    Exit a pool using the specified pool shares

    Attributes:
        sender (str): The creator address
        pool_id (int): The id of the pool
        pool_shares (Coin): The tokens as share of the pool to exit with
    """

    sender: str
    pool_id: int
    pool_shares: Coin

    def to_pb(self) -> pb.MsgExitPool:
        """
        Returns the Message as protobuf object.

        Returns:
            pb.MsgExitPool: The proto object.

        """
        return pb.MsgExitPool(
            sender=self.sender,
            pool_id=self.pool_id,
            pool_shares=self.pool_shares._generate_proto_object(),
        )


@dataclasses.dataclass
class MsgSwapAssets(PythonMsg):
    """
    Swap the assets provided for the denom specified

    Attributes:
        sender (str): The creator address
        pool_id (int): The id of the pool
        token_in (Coin): The token in we wish to swap with
        token_out_denom (str): The token we expect out of the pool
    """

    sender: str
    pool_id: int
    token_in: Coin
    token_out_denom: str

    def to_pb(self) -> pb.MsgSwapAssets:
        """
        Returns the Message as protobuf object.

        Returns:
            pb.MsgSwapAssets: The proto object.

        """
        return pb.MsgSwapAssets(
            sender=self.sender,
            pool_id=self.pool_id,
            token_in=self.token_in._generate_proto_object(),
            token_out_denom=self.token_out_denom,
        )


class spot:
    """
    The spot class allows to create transactions for the decentralized spot exchange using the queries.
    """

    create_pool: MsgCreatePool
    join_pool: MsgJoinPool
    exit_pool: MsgExitPool
    swap_assets: MsgSwapAssets



from grpc import Channel
from nibiru_proto.proto.vpool.v1 import query_pb2 as vpool_type
from nibiru_proto.proto.vpool.v1 import query_pb2_grpc as vpool_query
from nibiru_proto.proto.vpool.v1.state_pb2 import Direction as pbDirection

from nibiru.pytypes import Direction
from nibiru.query_clients.util import QueryClient


class VpoolQueryClient(QueryClient):
    """
    VPool allows to query the endpoints made available by the Nibiru Chain's VPOOL module.
    """

    def __init__(self, channel: Channel):
        self.api = vpool_query.QueryStub(channel)

    def reserve_assets(self, pair: str):
        """
        Returns the reserves of the vpool

        Example Return Value::

        ```json
        {
          "base_asset_reserve": "142621265.346706602199856348",
          "quote_asset_reserve": "2401396363770.998849167876702560"
        }
        ```

        Args:
            pair: the pair in "base:quote" format

        Returns:
            dict: the reserves of the vpool

        """
        req = vpool_type.QueryReserveAssetsRequest(pair=pair)
        return self.query(self.api.ReserveAssets, req)

    def all_pools(self):
        """
        Returns information about all the existing vpools

        Example Return Value::

        ```json
        {
            "pools": [
                {
                    "pair": "ubtc:unusd",
                    "base_asset_reserve": "142621265.346706602199856348",
                    "quote_asset_reserve": "2401396363770.998849167876702560",
                    "config": {
                        "trade_limit_ratio": "0.100000000000000000",
                        "fluctuation_limit_ratio": "0.100000000000000000",
                        "max_oracle_spread_ratio": "0.100000000000000000",
                        "maintenance_margin_ratio": "0.062500000000000000",
                        "max_leverage": "10.000000000000000000"
                    }
                },...
            ],
            "prices": [
                {
                    "pair": "ubtc:unusd",
                    "mark_price": "16837.575784601967733255",
                    "index_price": "16810.222187500000000000",
                    "twap_mark": "16843.132913821956433353",
                    "swap_invariant": "342490188000000000000",
                    "block_number": "432541"
                },
            ]
        }
        ```

        Returns:
            dict: the list of existing pools

        """
        req = vpool_type.QueryAllPoolsRequest()
        resp = self.query(self.api.AllPools, req)
        for _, prices in enumerate(resp["prices"]):
            prices["index_price"] = cast_str_to_float_safely(prices["index_price"])
            prices["twap_mark"] = cast_str_to_float_safely(prices["twap_mark"])
        return resp

    def base_asset_price(self, pair: str, direction: Direction, base_asset_amount: str):
        """
        Returns the price at which a base asset amount is valued in quote

        Example Return Value::

        ```json
        {
          "price_in_quote_denom": "16837.575682263963270039"
        }
        ```

        Args:
            pair: the pair in "base:quote" format
            direction(Direction): the direction of the operation (add or remove)
            base_asset_amount: the amount of base we want to know the price in quote

        Returns:
            dict: the price of the asset in quote
        """
        dir_pb = pbDirection.DIRECTION_UNSPECIFIED
        if direction == Direction.ADD:
            dir_pb = pbDirection.ADD_TO_POOL
        elif direction == Direction.REMOVE:
            dir_pb = pbDirection.REMOVE_FROM_POOL

        req = vpool_type.QueryBaseAssetPriceRequest(
            pair=pair, direction=dir_pb, base_asset_amount=base_asset_amount
        )
        return self.query(self.api.BaseAssetPrice, req)


def cast_str_to_float_safely(number_str: str) -> float:
    """
    Cast a string to float safely, if it fails it returns 0.

    Args:
        number_str: the number in string

    Returns:
        number(float): the number as float

    """
    try:
        number = float(number_str)
    except:
        number = float(0)
    return number



"""
The network class allows the user to defines the network the sdk interface should connect to.

There are some default values set for devnet, testnet, mainet and localnet, but the user cna also define its own
network by setting the values of the Network data class.

"""
import dataclasses
import enum
import os
from typing import Dict, List, Optional


@dataclasses.dataclass
class Network:
    """A representation of a Nibiru network based on its Tendermint RPC, gRPC,
    and LCD (REST) endpoints. A 'Network' instance enables interactions with a
    running blockchain.

    Attributes:
        lcd_endpoint (str): .
        grpc_endpoint (str): .
        tendermint_rpc_endpoint (str): .
        chain_id (str): .
        websocket_endpoint (str): .
        env (Optional[str]): TODO docs
        fee_denom (Optional[str]): Denom for the coin used to pay gas fees. Defaults to "unibi".

    Methods:
        customnet: A custom Nibiru network based on environment variables.
            Defaults to localnet.
        devnet: A development testnet environment that runs the latest release or
            pre-release from the nibiru repo. Defaults to 'nibiru-devnet-1'.
        localnet: The default local network created by running 'make localnet' in
            the nibiru repo.
        testnet: A stable testnet environment with public community members.
            Think of this as out practice mainnet. Defaults to 'nibiru-testnet-1'.
        mainnet: NotImplemented.

    Examples:
    >>> from nibiru import Network
    >>> network = Network.devnet(2)
    >>> network.is_insecure
    True
    """

    lcd_endpoint: str
    grpc_endpoint: str
    tendermint_rpc_endpoint: str
    chain_id: str
    websocket_endpoint: str
    env: str = "custom"
    fee_denom: str = "unibi"

    @property
    def is_insecure(self) -> bool:
        return not ("https" in self.tendermint_rpc_endpoint)

    @classmethod
    def customnet(cls) -> "Network":
        """
        Custom is the network configured from ENV variables.
        Defaults to localnet if no ENV variables are provided.

        Raises:
            KeyError: If the values are not set in the testing environment, this will raise an exception.

        Returns:
            Network: The updated Network object.
        """
        chain_config: Dict[str, Optional[str]] = {
            "LCD_ENDPOINT": os.getenv("LCD_ENDPOINT", "http://localhost:1317"),
            "GRPC_ENDPOINT": os.getenv("GRPC_ENDPOINT", "localhost:9090"),
            "TENDERMINT_RPC_ENDPOINT": os.getenv(
                "TENDERMINT_RPC_ENDPOINT", "http://localhost:26657"
            ),
            "WEBSOCKET_ENDPOINT": os.getenv(
                "WEBSOCKET_ENDPOINT", "ws://localhost:26657/websocket"
            ),
            "CHAIN_ID": os.getenv("CHAIN_ID", "nibiru-localnet-0"),
        }
        for name, env_var in chain_config.items():
            if env_var is None:
                raise KeyError(
                    "\n".join(
                        [
                            f"Environment variable {name} is needed for devnet.",
                            f"Please set {name} in your .env file.",
                        ]
                    )
                )

        return cls(
            lcd_endpoint=chain_config["LCD_ENDPOINT"],
            grpc_endpoint=chain_config["GRPC_ENDPOINT"],
            tendermint_rpc_endpoint=chain_config["TENDERMINT_RPC_ENDPOINT"],
            websocket_endpoint=chain_config["WEBSOCKET_ENDPOINT"],
            chain_id=chain_config["CHAIN_ID"],
            fee_denom='unibi',
            env="devnet",
        )

    @classmethod
    def testnet(cls, chain_num: int = 2) -> "Network":
        """
        Testnet is a network open to invited validators. It is more stable than
        devnet and provides a faucet to get some funds

        Args:
          chain_num (int): Testnet number

        Returns:
            Network: The updated Network object.
        """
        return cls(
            lcd_endpoint=f'https://lcd.testnet-{chain_num}.nibiru.fi',
            grpc_endpoint=f'tcp://grpc.testnet-{chain_num}.nibiru.fi:443',
            tendermint_rpc_endpoint=f'https://rpc.testnet-{chain_num}.nibiru.fi',
            websocket_endpoint=f'wss://rpc.testnet-{chain_num}.nibiru.fi/websocket',
            chain_id=f'nibiru-testnet-{chain_num}',
            fee_denom='unibi',
            env='testnet',
        )

    @classmethod
    def mainnet(cls) -> "Network":
        """
        Soon!
        """
        raise NotImplementedError

    @classmethod
    def localnet(cls) -> "Network":
        """
        Localnet is the network you would expect to connect to if you run `make localnet` from the nibiru repository.
        It allows you to update locally the golang codebase to checkout the behavior of the chain with different changes
        applied.

        Returns:
            Network: The updated Network object.
        """
        return cls(
            lcd_endpoint='http://localhost:1317',
            grpc_endpoint='localhost:9090',
            tendermint_rpc_endpoint='http://localhost:26657',
            websocket_endpoint='ws://localhost:26657/websocket',
            chain_id='nibiru-localnet-0',
            fee_denom='unibi',
            env='local',
        )

    def string(self) -> str:
        """
        Returns the current environment the network was initialized with. Will return `custom` if a custom network
        was created

        Returns:
            str: The name of the current environment.
        """
        return self.env

    @classmethod
    def devnet(cls, chain_num: int = 3) -> "Network":
        """
        Devnet is a network open to invited validators.

        Args:
          chain_num (int): Devnet number

        Returns:
            Network: The updated Network object.
        """
        return cls(
            lcd_endpoint=f'https://lcd.devnet-{chain_num}.nibiru.fi',
            grpc_endpoint=f'tcp://grpc.devnet-{chain_num}.nibiru.fi:443',
            tendermint_rpc_endpoint=f'https://rpc.devnet-{chain_num}.nibiru.fi:443',
            websocket_endpoint=f'wss://rpc.devnet-{chain_num}.nibiru.fi/websocket',
            chain_id=f'nibiru-devnet-{chain_num}',
            fee_denom='unibi',
            env='devnet',
        )

    @classmethod
    def from_chain_id(cls, chain_id: str) -> "Network":
        """
        Soon!
        """

        chain_id_elements: List[str] = chain_id.split("-")
        if len(chain_id_elements) != 3:
            raise ValueError(
                f"invalid chain_id format: {chain_id}. Expected one like nibiru-testnet-2"
            )

        prefix, chain_type, chain_number = chain_id_elements
        chain_number = int(chain_number)

        if chain_type == NetworkType.DEVNET.value:
            return Network.devnet(chain_number)
        elif chain_type == NetworkType.TESTNET.value:
            return Network.testnet(chain_number)
        elif chain_type == NetworkType.LOCALNET.value:
            return Network.localnet()
        else:
            network_types: List[str] = [member.value for member in NetworkType]
            raise ValueError(
                f"invalid chain type: {chain_type}. Available options: {network_types}"
            )


class NetworkType(enum.Enum):
    """Enum class for the available network types. E.g. 'testnet' and 'devnet'."""

    DEVNET = "devnet"
    TESTNET = "testnet"
    LOCALNET = "localnet"



from google.protobuf.json_format import MessageToDict
from grpc import Channel
from nibiru_proto.proto.cosmos.auth.v1beta1 import query_pb2 as auth_type
from nibiru_proto.proto.cosmos.auth.v1beta1 import query_pb2_grpc as auth_query

from nibiru.query_clients.util import QueryClient


class AuthQueryClient(QueryClient):
    """
    AuthQueryClient allows to query the endpoints made available by the Nibiru Chain's auth module.
    """

    def __init__(self, channel: Channel):
        self.api = auth_query.QueryStub(channel)

    def account(self, address: str) -> dict:
        """
        Returns information of the given account

        Example Return Value::

        ```json
        {
          "address": "nibi1zaavvzxez0elundtn32qnk9lkm8kmcsz44g7xl",
          "pub_key": null,
          "account_number": "17",
          "sequence": "0"
        }
        ```

        Args:
            address(str): the address of the account we want to get information

        Returns:
            dict: a dictionary containing information of the account

        """
        proto_output = self.query(
            api_callable=self.api.Account,
            req=auth_type.QueryAccountRequest(address=address),
            should_deserialize=False,
        )

        output = MessageToDict(proto_output)
        return output

    def accounts(self) -> dict:
        """
        Accounts returns all the existing accounts

        Example Return Value::

        ```json
        {
        "accounts": [
            {
              "address": "nibi1qqrg9ntgkavhxyn3zq0zmz85yp8vxgd9gm85pp",
              "pub_key": null,
              "account_number": "16246",
              "sequence": "0"
            },
            {
              "address": "nibi1qqrhxfwt6eu66v6mcdt7zc6xhqnw0w54u3rep3",
              "pub_key": null,
              "account_number": "6927",
              "sequence": "0"
            }
          ]
        }
        ```

        Returns:
            dict: a dictionary with information of all the existing accounts paginated

        """
        proto_output = self.query(
            api_callable=self.api.Accounts,
            req=auth_type.QueryAccountRequest(),
            should_deserialize=False,
        )

        output = MessageToDict(proto_output)
        return output



import dataclasses
from typing import List

from nibiru_proto.proto.perp.v1 import state_pb2 as state_pb
from nibiru_proto.proto.perp.v1 import tx_pb2 as pb

from nibiru.pytypes import Coin, PythonMsg, Side
from nibiru.utils import to_sdk_dec, to_sdk_int


@dataclasses.dataclass
class Liquidation:
    """
    Keeper of the pair/trader pairs for liquidations
    """

    pair: str
    trader: str


class MsgsPerp:
    """
    Messages for the Nibiru Chain x/perp module

    Methods:
    - open_position
    - close_position:
    - add_margin: Deleverages a position by adding margin to back it.
    - remove_margin: Increases the leverage of the position by removing margin.
    """

    @staticmethod
    def open_position(
        sender: str,
        pair: str,
        is_long: bool,
        quote_asset_amount: float,
        leverage: float,
        base_asset_amount_limit: float,
    ) -> 'MsgOpenPosition':
        """
        Open a posiiton using the specified parameters.

        Attributes:
            sender (str): The sender address
            pair (str): The token pair
            is_long (bool): Determines whether to open with long or short exposure.
            quote_asset_amount (float): The quote amount you want to use to buy base
            leverage (float): The leverage you want to use, typically between 1 and 15, depending on the maintenance
                margin ratio of the pool.
            base_asset_amount_limit (float): The minimum amount of base you are willing to receive for this amount of
                quote.
        """
        side: Side
        if is_long:
            side = Side.BUY
        else:
            side = Side.SELL
        return MsgOpenPosition(
            sender=sender,
            pair=pair,
            side=side,
            quote_asset_amount=quote_asset_amount,
            leverage=leverage,
            base_asset_amount_limit=base_asset_amount_limit,
        )

    @staticmethod
    def close_position(
        sender: str,
        pair: str,
    ) -> 'MsgClosePosition':
        """
        Close the position.

        Attributes:
            sender (str): The sender address
            pair (str): The token pair
        """
        return MsgClosePosition(sender=sender, pair=pair)

    @staticmethod
    def add_margin(
        sender: str,
        pair: str,
        margin: Coin,
    ) -> 'MsgAddMargin':
        """
        Add margin for the position (pair + trader)

        Attributes:
            sender (str): The trader address
            pair (str): The token pair
            margin (Coin): The margin to remove in a coin format
        """
        return MsgAddMargin(sender=sender, pair=pair, margin=margin)

    @staticmethod
    def remove_margin(
        sender: str,
        pair: str,
        margin: Coin,
    ) -> 'MsgRemoveMargin':
        """
        Remove margin for the position (pair + trader)

        Attributes:
            sender (str): The trader address
            pair (str): The token pair
            margin (Coin): The margin to remove in a coin format
        """
        return MsgRemoveMargin(sender=sender, pair=pair, margin=margin)

    @staticmethod
    def liquidate(
        sender: str,
        pair: str,
        trader: str,
    ) -> 'MsgMultiLiquidate':
        """
        Liquidates unhealthy position (pair + trader)

        Attributes:
            sender (str): The liquidator address
            pair (str): The token pair
            trader (str): The trader address
        """
        return MsgMultiLiquidate(
            sender=sender, liquidations=[Liquidation(pair=pair, trader=trader)]
        )

    @staticmethod
    def liquidate_multiple(
        sender: str, liquidations: List[Liquidation]
    ) -> 'MsgMultiLiquidate':
        """
        Liquidates multiple unhealthy positions (pair + trader)

        Attributes:
            sender (str): The liquidator address
            liquidations (List[Liquidation]): list of pair/traders to liquidate
        """
        return MsgMultiLiquidate(sender=sender, liquidations=liquidations)


@dataclasses.dataclass
class MsgRemoveMargin(PythonMsg):
    """
    Remove margin for the position (pair + trader)

    Attributes:
        sender (str): The trader address
        pair (str): The token pair
        margin (Coin): The margin to remove in a coin format
    """

    sender: str
    pair: str
    margin: Coin

    def to_pb(self) -> pb.MsgRemoveMargin:
        """
        Returns the Message as protobuf object.

        Returns:
            pb.MsgRemoveMargin: The proto object.

        """
        return pb.MsgRemoveMargin(
            sender=self.sender,
            pair=self.pair,
            margin=self.margin._generate_proto_object(),
        )


@dataclasses.dataclass
class MsgAddMargin(PythonMsg):
    """
    Add margin for the position (pair + trader)

    Attributes:
        sender (str): The trader address
        pair (str): The token pair
        margin (Coin): The margin to remove in a coin format
    """

    sender: str
    pair: str
    margin: Coin

    def to_pb(self) -> pb.MsgAddMargin:
        """
        Returns the Message as protobuf object.

        Returns:
            pb.MsgAddMargin: The proto object.

        """
        return pb.MsgAddMargin(
            sender=self.sender,
            pair=self.pair,
            margin=self.margin._generate_proto_object(),
        )


@dataclasses.dataclass
class MsgOpenPosition(PythonMsg):
    """
    Open a position using the specified parameters.

    Attributes:
        sender (str): The sender address
        pair (str): The token pair
        side (Side): The side, either Side.BUY or Side.SELL
        quote_asset_amount (float): The quote amount you want to use to buy base
        leverage (float): The leverage you want to use, typically between 1 and 15, depending on the maintenance
            margin ratio of the pool.
        base_asset_amount_limit (float): The minimum amount of base you are willing to receive for this amount of
            quote.
    """

    sender: str
    pair: str
    side: Side
    quote_asset_amount: float
    leverage: float
    base_asset_amount_limit: float

    def to_pb(self) -> pb.MsgOpenPosition:
        """
        Returns the Message as protobuf object.

        Returns:
            pb.MsgOpenPosition: The proto object.

        """
        pb_side = state_pb.Side.BUY if self.side == Side.BUY else state_pb.SELL
        quote_asset_amount_pb = to_sdk_int(self.quote_asset_amount)
        base_asset_amount_limit_pb = to_sdk_int(self.base_asset_amount_limit)
        leverage_pb = to_sdk_dec(self.leverage)

        return pb.MsgOpenPosition(
            sender=self.sender,
            pair=self.pair,
            side=pb_side,
            quote_asset_amount=quote_asset_amount_pb,
            leverage=leverage_pb,
            base_asset_amount_limit=base_asset_amount_limit_pb,
        )


@dataclasses.dataclass
class MsgClosePosition(PythonMsg):
    """
    Close the position.

    Attributes:
        sender (str): The sender address
        pair (str): The token pair
    """

    sender: str
    pair: str

    def to_pb(self) -> pb.MsgClosePosition:
        """
        Returns the Message as protobuf object.

        Returns:
            pb.MsgClosePosition: The proto object.

        """
        return pb.MsgClosePosition(
            sender=self.sender,
            pair=self.pair,
        )


@dataclasses.dataclass
class MsgMultiLiquidate(PythonMsg):
    """
    Close the position.

    Attributes:
        sender (str): The sender address
        liquidations (Liquidation): The list of {pair, trader} pairs.
    """

    sender: str
    liquidations: List[Liquidation]

    def to_pb(self) -> pb.MsgMultiLiquidate:
        """
        Returns the Message as protobuf object.

        Returns:
            pb.MsgLiquidate: The proto object.

        """
        return pb.MsgMultiLiquidate(
            sender=self.sender,
            liquidations=[
                pb.MsgMultiLiquidate.Liquidation(
                    pair=liquidation.pair,
                    trader=liquidation.trader,
                )
                for liquidation in self.liquidations
            ],
        )


class perp:
    """
    The perp class allows you to generate transaction for the perpetual futures module
    using the different messages available.
    """

    remove_margin: MsgRemoveMargin
    add_margin: MsgAddMargin
    open_position: MsgOpenPosition
    close_position: MsgClosePosition
    liquidate: MsgMultiLiquidate



from grpc import Channel
from nibiru_proto.proto.cosmos.staking.v1beta1 import query_pb2 as staking_type
from nibiru_proto.proto.cosmos.staking.v1beta1 import query_pb2_grpc as staking_query

from nibiru.query_clients.util import QueryClient, deserialize, get_page_request


class StakingQueryClient(QueryClient):
    """
    StakingQueryClient allows to query the endpoints made available
    by the Nibiru Chain's staking module.
    """

    def __init__(self, channel: Channel):
        self.api = staking_query.QueryStub(channel)

    def pool(self) -> dict:
        """
        Query the current staking pool values

        Example Return Value::

        ```json
        {
          "not_bonded_tokens": "49100355602",
          "bonded_tokens": "2419845520076726"
        }
        ```

        Returns:
            dict: the staking pool values by bounded and not bounded

        """
        return self.query(
            api_callable=self.api.Pool,
            req=staking_type.QueryPoolRequest(),
        )

    def delegation(self, delegator_addr: str, validator_addr: str) -> dict:
        """
        Query a delegation based on address and validator address

        Args:
            delegator_addr: the delegator address
            validator_addr: the validator that the delegator has delegated

        Returns:
            dict: the delegation details

        Example Return Value::

        ```json
        {
          "delegation": {
            "delegator_address": "nibi1zpsd7pxvqfcaxevgc2a07q3tgcwgaua92uyg4x",
            "validator_address": "nibivaloper1zpsd7pxvqfcaxevgc2a07q3tgcwgaua9r82npm",
            "shares": "127051284777.000000000000000000"
          },
          "balance": {
            "denom": "unibi",
            "amount": "127051284777"
          }
        }
        ```
        """
        return self.query(
            api_callable=self.api.Delegation,
            req=staking_type.QueryDelegationRequest(
                delegator_addr=delegator_addr, validator_addr=validator_addr
            ),
        )

    def delegations(self, delegator_addr: str, **kwargs) -> dict:
        """
        Query all delegations made by one delegator

        Example Return Value::

        ```json
        {
            "delegation_responses": [
                {
                    "delegation": {
                    "delegator_address": "nibi1zpsd7pxvqfcaxevgc2a07q3tgcwgaua92uyg4x",
                    "validator_address": "nibivaloper1zpsd7pxvqfcaxevgc2a07q3tgcwgaua9r82npm",
                    "shares": "127114301897.000000000000000000"
                },
                "balance": {
                    "denom": "unibi",
                    "amount": "127114301897"
                }
            }
        }
        ```

        Args:
            delegator_addr: the address of the delegator

        Returns:
            dict: the delegations that de delegator has
        """
        return self.query(
            api_callable=self.api.DelegatorDelegations,
            req=staking_type.QueryDelegatorDelegationsRequest(
                delegator_addr=delegator_addr,
                pagination=get_page_request(kwargs),
            ),
        )

    def delegations_to(self, validator_addr: str, **kwargs) -> dict:
        """
        Query all delegations made to one validator

        Return Example::

        ```json
        {
        "delegation_responses": [
          {
            "delegation": {
            "delegator_address": "nibi1zpsd7pxvqfcaxevgc2a07q3tgcwgaua92uyg4x",
            "validator_address": "nibivaloper1zpsd7pxvqfcaxevgc2a07q3tgcwgaua9r82npm",
            "shares": "127429922469.000000000000000000"
          },
          "balance": {
            "denom": "unibi",
            "amount": "127429922469"
          }
        ],
        }
        ```

        Args:
            validator_addr: the validator that we want to know the delegations it has

        Returns:
            dict: the list of delegations
        """
        return self.query(
            api_callable=self.api.ValidatorDelegations,
            req=staking_type.QueryValidatorDelegationsRequest(
                validator_addr=validator_addr,
                pagination=get_page_request(kwargs),
            ),
        )

    def historical_info(self, height: int) -> dict:
        """
        Query historical info at given height

        Args:
            height: at which we want to get historical info

        Returns:
            dict: the historical information
        """
        return self.query(
            api_callable=self.api.HistoricalInfo,
            req=staking_type.QueryHistoricalInfoRequest(height=height),
        )

    def params(self) -> dict:
        """
        Query the current staking parameters information

        Example Return Value::

        ```json
        {
          "unbonding_time": "7200s",
          "max_validators": 100,
          "max_entries": 7,
          "historical_entries": 10000,
          "bond_denom": "unibi"
        }
        ```

        Returns:
            dict: the params of the module
        """
        return self.query(
            api_callable=self.api.Params,
            req=staking_type.QueryParamsRequest(),
        )

    def redelegations(
        self, delegator_addr: str, dst_validator_addr: str, **kwargs
    ) -> dict:
        """
        Query all redelegations records for one delegator

        Args:
            delegator_addr: the address of the delegator
            dst_validator_addr: a validator address
        """
        return self.query(
            api_callable=self.api.Redelegations,
            req=staking_type.QueryRedelegationsRequest(
                delegator_addr=delegator_addr,
                dst_validator_addr=dst_validator_addr,
                pagination=get_page_request(kwargs),
            ),
        )

    def unbonding_delegation(self, delegator_addr: str, validator_addr: str) -> dict:
        """
        Query an unbonding-delegation record based on delegator and validator address

        Args:
            delegator_addr: the delegator address
            validator_addr: the validator address
        """
        return self.query(
            api_callable=self.api.UnbondingDelegation,
            req=staking_type.QueryUnbondingDelegationRequest(
                validator_addr=validator_addr,
                delegator_addr=delegator_addr,
            ),
        )

    def unbonding_delegations(self, delegator_addr: str, **kwargs) -> dict:
        """
        Query all unbonding-delegations records for one delegator

        Args:
            delegator_addr: the delegator address
        """
        return self.query(
            api_callable=self.api.DelegatorUnbondingDelegations,
            req=staking_type.QueryDelegatorUnbondingDelegationsRequest(
                delegator_addr=delegator_addr,
                pagination=get_page_request(kwargs),
            ),
        )

    def unbonding_delegations_from(self, validator_addr: str, **kwargs) -> dict:
        """
        Query all unbonding delegations from a validator

        Args:
            validator_addr: the validator address
        """
        return self.query(
            api_callable=self.api.ValidatorUnbondingDelegations,
            req=staking_type.QueryValidatorUnbondingDelegationsRequest(
                validator_addr=validator_addr,
                pagination=get_page_request(kwargs),
            ),
        )

    def validator(self, validator_addr: str) -> dict:
        """
        Query a validator

        Example Return Value::

        ```json
        {
          "operator_address": "nibivaloper1zpsd7pxvqfcaxevgc2a07q3tgcwgaua9r82npm",
          "consensus_pubkey": {
            "@type": "/cosmos.crypto.ed25519.PubKey",
            "key": "9hFIS+BTJxyWH/OgjH4om2nOWf2p7whS/jmMuEA4gnk="
          },
          "jailed": false,
          "status": "BOND_STATUS_BONDED",
          "tokens": "127702800941",
          "delegator_shares": "127702800941.000000000000000000",
          "description": {
            "moniker": "some val",
            "identity": "info",
            "website": "someweb",
            "security_contact": "",
            "details": "abc"
          },
          "unbonding_height": "0",
          "unbonding_time": "1970-01-01T00:00:00Z",
          "commission": {
            "commission_rates": {
              "rate": "0.100000000000000000",
              "max_rate": "0.300000000000000000",
              "max_change_rate": "0.030000000000000000"
            },
            "update_time": "2022-12-21T17:17:38.806603554Z"
          },
          "min_self_delegation": "1"
        }
        ```

        Args:
            validator_addr: the validator address

        Returns:
            dict: the information from the validator
        """
        return self.query(
            api_callable=self.api.Validator,
            req=staking_type.QueryValidatorRequest(
                validator_addr=validator_addr,
            ),
        )

    def validators(self, **kwargs) -> dict:
        """
        Query for all validators

        Example Return Value::
            Same as the return value of the 'validator' function except as a list

        Returns:
            dict: the info of all validators

        """
        proto_output = self.query(
            api_callable=self.api.Validators,
            req=staking_type.QueryValidatorsRequest(
                pagination=get_page_request(kwargs),
            ),
            should_deserialize=False,
            height=kwargs.get("height"),
        )
        return deserialize(proto_output)



import dataclasses
from typing import Iterable, List, Union

from nibiru_proto.proto.cosmos.bank.v1beta1 import tx_pb2 as pb

from nibiru.pytypes import Coin, PythonMsg


class MsgsBank:
    """
    Messages for the x/bank module.

    Methods:
    - send: Send tokens from one account to another
    """

    @staticmethod
    def send(
        from_address: str,
        to_address: str,
        coins: Union[Coin, List[Coin]],
    ) -> 'MsgSend':
        """
        Send tokens from one account to another

        Attributes:
            from_address (str): The address of the sender
            to_address (str): The address of the receiver
            coins (List[Coin]): The list of coins to send

        Returns:
            MsgSend: PythonMsg corresponding to the 'cosmos.bank.v1beta1.MsgSend' message
        """
        return MsgSend(from_address=from_address, to_address=to_address, coins=coins)


@dataclasses.dataclass
class MsgSend(PythonMsg):
    """
    Send tokens from one account to another. PythonMsg corresponding to the
    'cosmos.bank.v1beta1.MsgSend' message.

    Attributes:
        from_address (str): The address of the sender
        to_address (str): The address of the receiver
        coins (Union[Coin, List[Coin]]): The list of coins to send
    """

    from_address: str
    to_address: str
    coins: Union[Coin, List[Coin]]

    def to_pb(self) -> pb.MsgSend:
        """
        Returns the Message as protobuf object.

        Returns:
            pb.MsgSend: The proto object.

        """
        coins = self.coins
        if not isinstance(coins, Iterable):
            coins = [self.coins]

        return pb.MsgSend(
            from_address=self.from_address,
            to_address=self.to_address,
            amount=[coin._generate_proto_object() for coin in coins],
        )



from typing import Dict, List, Union

from google.protobuf.json_format import MessageToDict
from grpc import Channel
from nibiru_proto.proto.perp.v1 import query_pb2 as perp_type
from nibiru_proto.proto.perp.v1 import query_pb2_grpc as perp_query

from nibiru.query_clients.util import QueryClient, deserialize
from nibiru.utils import from_sdk_dec


class PerpQueryClient(QueryClient):
    """
    Perp allows to query the endpoints made available by the Nibiru Chain's PERP module.
    """

    def __init__(self, channel: Channel):
        self.api = perp_query.QueryStub(channel)

    def params(self):
        """
        Get the parameters of the perp module.

        Example Return Value::

        ```json
        {
          "feePoolFeeRatio": 0.001,
          "ecosystemFundFeeRatio": 0.001,
          "liquidationFeeRatio": 0.025,
          "partialLiquidationRatio": 0.25,
          "epochIdentifier": "30 min",
          "twapLookbackWindow": "900s"
        }
        ```

        Returns:
            dict: The current parameters for the perpetual module
        """
        proto_output: perp_type.QueryParamsResponse = self.query(
            api_callable=self.api.Params,
            req=perp_type.QueryParamsRequest(),
            should_deserialize=False,
        )

        output = MessageToDict(proto_output)["params"]

        sdk_dec_fields = [
            "feePoolFeeRatio",
            "ecosystemFundFeeRatio",
            "liquidationFeeRatio",
            "partialLiquidationRatio",
        ]

        for field in sdk_dec_fields:
            output[field] = from_sdk_dec(output[field])

        return output

    def position(self, pair: str, trader: str) -> dict:
        """
        Get the trader position. Returns information about position notional, margin ratio
        unrealized pnl, size of the position etc.

        Args:
            pair (str): The token pair
            trader (str): The trader address

        Example Return Value::

        ```json
        {
          "position": {
            "traderAddress": "nibi1zaavvzxez0elund",
            "pair": "ubtc:unusd",
            "size": 11.241446725317692,
            "margin": 45999.99999999999,
            "openNotional": 230000.0,
            "lastUpdateCumulativePremiumFraction": "0",
            "blockNumber": "278"
          },
          "positionNotional": 230000.0,
          "unrealizedPnl": 1.024e-20,
          "marginRatioMark": 0.2,
          "marginRatioIndex": 0.2
        }
        ```

        Returns:
            dict: The output of the query
        """
        req = perp_type.QueryPositionRequest(
            pair=pair,
            trader=trader,
        )

        proto_output: perp_type.QueryPositionResponse = self.query(
            api_callable=self.api.QueryPosition, req=req, should_deserialize=False
        )

        return deserialize(proto_output)

    def all_positions(self, trader: str) -> Dict[str, dict]:
        """
        Args:
            trader (str): Address of the owner of the positions

        Returns:
            Dict[str, dict]: All of the open positions for the 'trader'.

        Example Return Value:

        ```json
        {
          "ubtc:unusd": {
            "block_number": 1137714,
            "margin_ratio_index": 0.0,
            "margin_ratio_mark": 0.09999999999655101,
            "position": {
            "block_number": 1137714,
            "latest_cumulative_premium_fraction": 17233.436302191654,
            "margin": 10.0,
            "open_notional": 100.0,
            "pair": "ubtc:unusd",
            "size": -0.00545940925278242,
            "trader_address": "nibi10gm4kys9yyrlqpvj05vqvjwvje87gln8nsm8wa"
            },
            "position_notional": 100.0,
            "unrealized_pnl": -5.079e-15
          }
        }
        ```
        """
        req = perp_type.QueryPositionsRequest(
            trader=trader,
        )

        proto_output: perp_type.QueryPositionsResponse = self.query(
            api_callable=self.api.QueryPositions, req=req, should_deserialize=False
        )
        proto_as_dict: dict[str, list] = deserialize(proto_output)

        position_resps: Union[List[dict], None] = proto_as_dict.get("positions")
        if position_resps is None:
            return proto_as_dict

        positions_map: Dict[str, dict] = {}
        for position_resp in position_resps:
            pair = position_resp["position"]["pair"]
            positions_map[pair] = position_resp

        return positions_map



from nibiru.query_clients.auth import AuthQueryClient  # noqa
from nibiru.query_clients.epoch import EpochQueryClient  # noqa
from nibiru.query_clients.perp import PerpQueryClient  # noqa
from nibiru.query_clients.spot import SpotQueryClient  # noqa
from nibiru.query_clients.stablecoin import StablecoinQueryClient  # noqa
from nibiru.query_clients.staking import StakingQueryClient  # noqa
from nibiru.query_clients.util import *  # noqa
from nibiru.query_clients.vpool import VpoolQueryClient  # noqa



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nibiru',
 'nibiru.crypto',
 'nibiru.msg',
 'nibiru.pytypes',
 'nibiru.query_clients']

package_data = \
{'': ['*']}

install_requires = \
['aiocron>=1.8,<2.0',
 'aiohttp>=3.8.3,<4.0.0',
 'asyncio>=3.4.3,<4.0.0',
 'bech32>=1.2.0,<2.0.0',
 'bip32>=3.3,<4.0',
 'ecdsa>=0.18.0,<0.19.0',
 'hdwallets>=0.1.2,<0.2.0',
 'importlib-metadata>=5.0.0,<6.0.0',
 'mnemonic>=0.20,<0.21',
 'nibiru-proto==0.19.0',
 'packaging>=21.3,<22.0',
 'pre-commit>=2.20.0,<3.0.0',
 'python-dotenv>=0.21.0,<0.22.0',
 'requests>=2.28.1,<3.0.0',
 'shutup>=0.2.0,<0.3.0',
 'websocket-client>=1.4.1,<2.0.0']

setup_kwargs = {
    'name': 'nibiru',
    'version': '0.19.2',
    'description': 'Python SDK for interacting with Nibiru.',
    'long_description': '# Python SDK - Nibiru Chain    <!-- omit in toc -->\n\n> Python-based client for interacting with the Nibiru blockchain.\n\n<!-- Badges -->\n\n[![Nibiru Test workflow][tests-badge]][tests-workflow]\n[![Nibiru examples tests][examples-badge]][tests-example]\n[![PyPI Version][pypi-image]][pypi-url]\n[![][documentation-image]][documentation-url]\n[![][discord-badge]][discord-url]\n[![][stars-image]][stars-url]\n[![MIT license][license-badge]][license-link]\n\n<!-- Badges links -->\n\n<!-- pypi -->\n[pypi-image]: https://img.shields.io/pypi/v/nibiru\n[pypi-url]: https://pypi.org/project/nibiru/\n[stars-image]: https://img.shields.io/github/stars/NibiruChain?style=social\n[stars-url]: https://github.com/NibiruChain\n[documentation-image]: https://readthedocs.org/projects/nibiru-py/badge/?version=latest\n[documentation-url]: https://nibiru-py.readthedocs.io/en/latest/?badge=latest\n[discord-badge]: https://dcbadge.vercel.app/api/server/sgPw8ZYfpQ?style=flat\n[discord-url]: https://discord.gg/sgPw8ZYfpQ\n[license-badge]: https://img.shields.io/badge/License-MIT-blue.svg\n[license-link]: https://github.com/NibiruChain/py-sdk/blob/master/LICENSE\n[tests-badge]: https://github.com/NibiruChain/py-sdk/actions/workflows/pytests.yml/badge.svg\n[examples-badge]: https://github.com/NibiruChain/py-sdk/actions/workflows/notebooks.yml/badge.svg\n[tests-workflow]: https://github.com/NibiruChain/py-sdk/actions/workflows/pytests.yml\n[tests-example]: https://github.com/NibiruChain/py-sdk/actions/workflows/notebooks.yml\n\nThe `nibiru` package allows you to index, query, and send transactions on Nibiru Chain using Python. It provides access to market data for analysis, visualization, indicator development, algorithmic trading, strategy backtesting, bot programming, and related software engineering.\n\nThe package is intended to be used by coders, developers, technically-skilled traders and  data-scientists for building trading algorithms.\n\n#### README Contents\n\n- [Python SDK Tutorial](#python-sdk-tutorial)\n- [Installation from `PyPI`](#installation-from-pypi)\n- [Usage](#usage)\n  - [Ex: Creating a wallet and SDK client](#ex-creating-a-wallet-and-sdk-client)\n  - [Ex: Using the faucet](#ex-using-the-faucet)\n  - [Ex: Querying chain state](#ex-querying-chain-state)\n  - [Ex: Submitting transactions](#ex-submitting-transactions)\n- [Documentation Website](#documentation-website)\n- [Contributing](#contributing)\n\n## Python SDK Tutorial\n\n<a href="https://colab.research.google.com/github/NibiruChain/py-sdk/blob/master/examples/collab_notebook.ipynb" target="_blank">\n<p align="center">\n  <img src="https://colab.research.google.com/assets/colab-badge.svg" style="width: 300px;">\n</p>\n</a>\n\n## Installation from `PyPI`\n\n```bash\npip install nibiru  # requires Python 3.7+\n```\n\nYou may need to update `pip` to get this to run:\n\n```bash\npython -m pip install --upgrade pip\n```\n\n## Usage\n\n### Ex: Creating a wallet and SDK client\n\n```python\nfrom nibiru import wallet\n\n# Save the mnemonic for later\nmnemonic, private_key = wallet.PrivateKey.generate()\n```\n\nAfter, creating an account, you can create an `Sdk` instance.\n\n```python\nimport nibiru\n\nnetwork = nibiru.network.Network.testnet(2)\nsdk = nibiru.Sdk.authorize(mnemonic)\n  .with_network(network)\n```\n\nThe `Sdk` class creates an interface to sign and send transactions or execute\nqueries. It is associated with:\n- A transaction signer (wallet), which is configured from existing mnemonic to recover a `PrivateKey`.\n- A `Network`, which specifies the RPC, LCD, and gRPC endpoints for connecting to Nibiru Chain.\n- An optional `TxConfig` for changing gas parameters.\n\n### Ex: Using the faucet\n\n```python\nimport requests\n\nrequests.post(\n    "https://faucet.testnet-2.nibiru.fi/",\n    json={\n        "address": sdk.address,\n        "coins": ["10000000unibi", "100000000000unusd"],\n    },\n)\n```\n\n### Ex: Querying chain state\n\n```python\n# Querying the token balances of the account\nsdk.query.get_bank_balances(sdk.address)\n\n# Querying from the vpool module\nquery_resp = sdk.query.vpool.all_pools()\nprint(query_resp)\n# Queries from other modules can be accessed from "sdk.query.module"\n```\n\n### Ex: Submitting transactions\n\n```python\n# version 0.16.3\nfrom nibiru import Msg\n\ntx_resp = sdk.tx.execute_msgs(\n    Msg.perp.open_position(\n        sender=sdk.address,\n        pair="ubtc:unusd",\n        is_long=True,\n        quote_asset_amount=10,\n        leverage=10,\n        base_asset_amount_limit=0,\n    )\n)\n```\n\nYou can broadcast any available transaction by passing its corresponding `Msg` to the `sdk.tx.execute_msgs` function.\n\n## Documentation Website\n\nDocumentation can be found here: [Nibiru-py documentation](https://nibiru-py.readthedocs.io/en/latest/index.html)\n\n- Learn more about opening and managing your spot and perp positions [here](https://nibiru-py.readthedocs.io/en/latest/nibiru.sdks.tx.html#nibiru-sdks-tx-package)\n- Learn about querying the chain using the Sdk [here](https://nibiru-py.readthedocs.io/en/latest/nibiru.clients.html#nibiru-clients-package)\n\n## Contributing\n\nPlease read [HACKING.MD](HACKING.md) for developer environment setup.\n',
    'author': 'Nibiru Chain',
    'author_email': 'dev@nibiru.fi',
    'maintainer': 'NibiruHeisenberg',
    'maintainer_email': 'dev@nibiru.fi',
    'url': 'https://github.com/NibiruChain/py-sdk',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from google.protobuf.json_format import MessageToDict
from grpc import Channel
from nibiru_proto.proto.epochs import query_pb2 as epoch_type
from nibiru_proto.proto.epochs import query_pb2_grpc as epoch_query

from nibiru.query_clients.util import QueryClient


class EpochQueryClient(QueryClient):
    """
    EpochQueryClient allows to query the endpoints made available by the epoch
    module of Nibiru Chain. This module is used to time out certain events like
    funding rate payments.
    """

    def __init__(self, channel: Channel):
        self.api = epoch_query.QueryStub(channel)

    def current_epoch(self, epoch_identifier: str) -> dict:
        """
        Returns information about the epoch specified

        Example Return Value::

        ```json
        {"current_epoch":"329"}
        ```

        Args:
            epoch_identifier(str): the identifier of the epoch, example: "week"

        Returns:

        """
        proto_output = self.query(
            api_callable=self.api.CurrentEpoch,
            req=epoch_type.QueryCurrentEpochRequest(identifier=epoch_identifier),
            should_deserialize=False,
        )

        output = MessageToDict(proto_output)
        output["currentEpoch"] = int(output["currentEpoch"])
        return output

    def epoch_infos(self) -> dict:
        """
        Returns all the epochs that exist and its details

        Example Return Value::

        ```json
        {
            "epochs": [
                {
                  "identifier": "15 min",
                  "start_time": "2022-12-16T22:55:34.903063726Z",
                  "duration": "900s",
                  "current_epoch": "658",
                  "current_epoch_start_time": "2022-12-23T19:10:34.903063726Z",
                  "epoch_counting_started": true,
                  "current_epoch_start_height": "427359"
                },
                {
                  "identifier": "30 min",
                  "start_time": "2022-12-16T22:55:34.903063726Z",
                  "duration": "1800s",
                  "current_epoch": "329",
                  "current_epoch_start_time": "2022-12-23T18:55:34.903063726Z",
                  "epoch_counting_started": true,
                  "current_epoch_start_height": "426741"
                }]
        }
        ```

        Returns:
            dict: the list of epochs
        """
        proto_output = self.query(
            api_callable=self.api.EpochInfos,
            req=epoch_type.QueryEpochsInfoRequest(),
            should_deserialize=False,
        )
        output = MessageToDict(proto_output)
        return output



from google.protobuf.json_format import MessageToDict
from grpc import Channel
from nibiru_proto.proto.stablecoin import query_pb2 as stablecoin_type
from nibiru_proto.proto.stablecoin import query_pb2_grpc as stablecoin_query

from nibiru.query_clients.util import QueryClient, dict_keys_from_camel_to_snake
from nibiru.utils import format_fields_nested, from_sdk_dec_n


class StablecoinQueryClient(QueryClient):
    """
    Stablecoin allows to query the endpoints made available by the Nibiru Chain's
    Stablecoin module.
    """

    def __init__(self, channel: Channel):
        self.api = stablecoin_query.QueryStub(channel)

    def params(self) -> dict:
        """
        Requests the parameters of the stablecoin module.

        Example Return Value:
        {
            "params": {
            "coll_ratio": "140000",
            "fee_ratio": "2000",
            "ef_fee_ratio": "500000",
            "bonus_rate_recoll": "2000",
            "distr_epoch_identifier": "15 min",
            "adjustment_step": "2500",
            "price_lower_bound": "999900",
            "price_upper_bound": "1000100",
            "is_collateral_ratio_valid": true
            }
        }

        Returns:
            dict: The parameters fo the stablecoin module.
        """
        proto_output = self.query(
            api_callable=self.api.Params,
            req=stablecoin_type.QueryParamsRequest(),
            should_deserialize=False,
        )
        output = MessageToDict(proto_output, including_default_value_fields=True)
        return dict_keys_from_camel_to_snake(
            format_fields_nested(
                object=output,
                fn=lambda x: from_sdk_dec_n(x, 6),
                fields=[
                    "collRatio",
                    "feeRatio",
                    "efFeeRatio",
                    "bonusRateRecoll",
                    "adjustmentStep",
                    "priceLowerBound",
                    "priceUpperBound",
                ],
            )
        )

    def circulating_supplies(self, **kwargs):
        """
        Return circulating supplies value of the stablecoin.

        Example Return Value::
        {
          "nibi": {
            "denom": "unibi",
            "amount": "4067118588898816"
          },
          "nusd": {
            "denom": "unusd",
            "amount": "2009322333444555"
          }
        }
        Returns:
            dict: The output of the query
        """
        proto_output = self.query(
            api_callable=self.api.CirculatingSupplies,
            req=stablecoin_type.QueryCirculatingSupplies(),
            should_deserialize=False,
        )
        output = MessageToDict(proto_output)
        return format_fields_nested(
            object=output, fn=lambda x: from_sdk_dec_n(x, 6), fields=["amount"]
        )

    def liquidity_ratio_info(self) -> dict:
        """
        Returns liquidity ratio from the stablecoin module

        Example Return Value:
        {
          "info": {
            "liquidity_ratio": "0.005214231356709351",
            "upper_band": "0.005219445588066060",
            "lower_band": "0.005209017125352642"
          }
        }

        Returns:
            dict: liquidity ratio info.
        """
        return self.query(
            api_callable=self.api.LiquidityRatioInfo,
            req=stablecoin_type.QueryLiquidityRatioInfoRequest(),
        )



import importlib
from typing import Dict, List, Optional, Union

import grpc
import grpc._channel
from google.protobuf import json_format, message
from grpc import Channel
from nibiru_proto.proto.cosmos.base.query.v1beta1.pagination_pb2 import PageRequest
from nibiru_proto.proto.cosmos.tx.v1beta1.tx_pb2 import Tx
from nibiru_proto.proto.tendermint.types.block_pb2 import Block
from nibiru_proto.proto.util.v1 import query_pb2 as util_type
from nibiru_proto.proto.util.v1 import query_pb2_grpc as util_query

from nibiru import utils
from nibiru.exceptions import QueryError

PROTOBUF_MSG_BASE_ATTRS: List[str] = (
    dir(message.Message)
    + ['Extensions', 'FindInitializationErrors', '_CheckCalledFromGeneratedFile']
    + ['_extensions_by_name', '_extensions_by_number']
)
"""PROTOBUF_MSG_BASE_ATTRS (List[str]): The default attributes and methods of
an instance of the 'protobuf.message.Message' class.
"""


def camel_to_snake(camel: str):
    return ''.join(
        ['_' + char.lower() if char.isupper() else char for char in camel]
    ).lstrip('_')


def dict_keys_from_camel_to_snake(d):
    """
    Transform all keys from the dictionary from camelcase to snake case.

    Args:
        d (dict): The dictionary to transform

    Returns:
        dict: The dictionary transformed
    """
    if isinstance(d, list):
        return [
            dict_keys_from_camel_to_snake(i) if isinstance(i, (dict, list)) else i
            for i in d
        ]
    return {
        camel_to_snake(a): dict_keys_from_camel_to_snake(b)
        if isinstance(b, (dict, list))
        else b
        for a, b in d.items()
    }


def deserialize(pb_msg: message.Message, no_sdk_transformation: bool = False) -> dict:
    """Deserializes a proto message into a dictionary.

    - sdk.Dec values are converted to floats.
    - sdk.Int values are converted to ints.
    - Missing fields become blank strings.

    Args:
        pb_msg (protobuf.message.Message)
        no_sdk_transformation (bool): Wether to bypass the sdk transformation. Default to False

    Returns:
        dict: 'pb_msg' as a JSON-able dictionary.
    """
    if not isinstance(pb_msg, message.Message):
        return pb_msg
    custom_dtypes: Dict[str, bytes] = {
        str(field[1]): field[0].GetOptions().__getstate__().get("serialized", None)
        for field in pb_msg.ListFields()
    }
    serialized_output = {}
    expected_fields: List[str] = list(pb_msg.DESCRIPTOR.fields_by_name.keys())

    for _, attr in enumerate(expected_fields):
        attr_search = pb_msg.__getattribute__(attr)
        custom_dtype = custom_dtypes.get(str(attr_search))

        if custom_dtype is not None:
            if "sdk/types.Dec" in str(custom_dtype):
                if no_sdk_transformation:
                    serialized_output[str(attr)] = float(pb_msg.__getattribute__(attr))
                else:
                    serialized_output[str(attr)] = utils.from_sdk_dec(
                        pb_msg.__getattribute__(attr)
                    )
            elif "sdk/types.Int" in str(custom_dtype):
                if no_sdk_transformation:
                    serialized_output[str(attr)] = int(pb_msg.__getattribute__(attr))
                else:
                    serialized_output[str(attr)] = utils.from_sdk_int(
                        pb_msg.__getattribute__(attr)
                    )
            elif "Int" in str(custom_dtype):  # Used for sdk.Coin message normalization
                serialized_output[str(attr)] = int(pb_msg.__getattribute__(attr))
            else:
                val = pb_msg.__getattribute__(attr)
                if hasattr(val, '__len__') and not isinstance(val, str):
                    updated_vals = []
                    for v in val:
                        updated_vals.append(deserialize(v))
                    serialized_output[str(attr)] = updated_vals
                else:
                    serialized_output[str(attr)] = deserialize(val)
        elif custom_dtype is None and not attr_search:
            if str(attr_search) == "[]":
                serialized_output[str(attr)] = []
            else:
                serialized_output[str(attr)] = attr_search
        else:
            serialized_output[str(attr)] = deserialize(pb_msg.__getattribute__(attr))

    return serialized_output


def deserialize_exp(proto_message: message.Message) -> dict:
    """
    Take a proto message and convert it into a dictionnary.
    sdk.Dec values are converted to be consistent with txs.

    Args:
        proto_message (protobuf.message.Message)

    Returns:
        dict
    """
    output = json_format.MessageToDict(proto_message)

    is_sdk_dec = {
        field.camelcase_name: "types.Dec" in str(field.GetOptions())
        for field in proto_message.DESCRIPTOR.fields
    }

    for field in proto_message.DESCRIPTOR.fields:
        if field.message_type is not None:
            # This is another proto object
            try:
                output[field.camelcase_name] = deserialize_exp(
                    proto_message.__getattribute__(field.camelcase_name)
                )
            except AttributeError:
                output[field.camelcase_name] = output[field.camelcase_name]

        elif is_sdk_dec[field.camelcase_name]:
            output[field.camelcase_name] = utils.from_sdk_dec(
                output[field.camelcase_name]
            )

    return dict_keys_from_camel_to_snake(output)


class QueryClient:
    def query(
        self,
        api_callable: grpc.UnaryUnaryMultiCallable,
        req: message.Message,
        should_deserialize: bool = True,
        height: Optional[int] = None,
    ) -> Union[dict, message.Message]:
        try:
            output: message.Message = api_callable(
                req,
                **(
                    {"metadata": (('x-cosmos-block-height', str(height)),)}
                    if height
                    else {}
                ),
            )
            if should_deserialize:
                return deserialize(output)
            return output
        except grpc._channel._InactiveRpcError as err:
            raise QueryError(
                f"Error on {str(api_callable._method).split('/')[-1][:-1]}: {err._state.details}"
            ) from None


def get_page_request(kwargs):
    return PageRequest(
        key=kwargs.get("key"),
        offset=kwargs.get("offset"),
        limit=kwargs.get("limit"),
        count_total=kwargs.get("count_total"),
        reverse=kwargs.get("reverse"),
    )


def get_msg_pb_by_type_url(type_url: str) -> Optional[message.Message]:
    """
    Tries loading protobuf class by type url.
    Examples type urls:
        /cosmos.bank.v1beta1.MsgSend
        /nibiru.perp.v1.MsgOpenPosition
    """
    class_ = None
    try:
        type_url = type_url.replace("/", "")
        module_name, class_name = type_url.rsplit(".", 1)
        if module_name.startswith("nibiru."):
            module_name = module_name.split(".", 1)[1]
        module_name = f"nibiru_proto.proto.{module_name}.tx_pb2"
        module_ = importlib.import_module(module_name)
        class_ = getattr(module_, class_name)()
    except Exception:
        pass

    return class_ or None


def get_block_messages(block: Block) -> List[dict]:
    """
    Rerurns block messages as a list of dicts.
    Matches corresponding messages types by type_url.
    """
    messages = []
    for tx in block.data.txs:
        tx_pb = Tx()
        tx_pb.MergeFromString(tx)

        for msg in tx_pb.body.messages:
            msg_pb: message.Message = get_msg_pb_by_type_url(msg.type_url)
            if msg_pb:
                msg_pb.ParseFromString(msg.value)
                messages.append(
                    {"type_url": msg.type_url, "value": deserialize(msg_pb)}
                )
    return messages


class UtilQueryClient(QueryClient):
    """
    UtilQueryClient allows to query the endpoints made available by the util
    module of Nibiru Chain.
    """

    def __init__(self, channel: Channel):
        self.api = util_query.QueryStub(channel)

    def module_accounts(self) -> dict:
        """
        Returns information module accounts like perp, vault etc.
        Includes account name, address and balances.

        Example Return Value:

        ```json
        {
          "accounts": [
            {
              "name": "perp",
              "address": "nibi1sr9a7yav4nu9n335atnkwgpcwz7s3p8puwaxgq",
              "balance": []
            },
            {
              "name": "vault",
              "address": "nibi1umc2r7a58jy3jmw0e0hctyy0rx45chmurptawl",
              "balance": []
            }
          ]
        }
        ```
        """
        return self.query(
            api_callable=self.api.ModuleAccounts,
            req=util_type.QueryModuleAccountsRequest(),
            should_deserialize=True,
        )
